# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Parse inference model predictions to triton inference responses
per model type.
"""

from functools import wraps
from itertools import zip_longest
from typing import Callable, Dict
import numpy as np

try:
  import triton_python_backend_utils as pb_utils
except ModuleNotFoundError:
  pass


def visual_detector(func: Callable):
  """
  Visual detector type output parser.
  """

  @wraps(func)
  def parse_predictions(self, input_data: np.ndarray, *args, **kwargs):
    """
    Format predictions and return clarifai compatible output.
    """
    out_bboxes = []
    out_labels = []
    out_scores = []

    # input_data passed as list of images
    preds = func(self, list(input_data[:]), *args, **kwargs)
    for pred in preds:
      out_bboxes.append(pred.predicted_bboxes)
      out_labels.append(pred.predicted_labels)
      out_scores.append(pred.predicted_scores)

    if len(out_bboxes) < 1 or len(out_labels) < 1:
      out_tensor_bboxes = pb_utils.Tensor("predicted_bboxes", np.zeros((0, 4), dtype=np.float32))
      out_tensor_labels = pb_utils.Tensor("predicted_labels", np.zeros((0, 1), dtype=np.int32))
      out_tensor_scores = pb_utils.Tensor("predicted_scores", np.zeros((0, 1), dtype=np.float32))
    else:
      out_tensor_bboxes = pb_utils.Tensor("predicted_bboxes",
                                          np.asarray(out_bboxes, dtype=np.float32))
      out_tensor_labels = pb_utils.Tensor("predicted_labels",
                                          np.asarray(out_labels, dtype=np.int32))
      out_tensor_scores = pb_utils.Tensor("predicted_scores",
                                          np.asarray(out_scores, dtype=np.float32))

    inference_response = pb_utils.InferenceResponse(
        output_tensors=[out_tensor_bboxes, out_tensor_labels, out_tensor_scores])

    return inference_response

  return parse_predictions


def visual_classifier(func: Callable):
  """
  Visual classifier type output parser.
  """

  @wraps(func)
  def parse_predictions(self, input_data: np.ndarray, *args, **kwargs):
    """
    Format predictions and return clarifai compatible output.
    """
    out_scores = []
    # input_data passed as list of images
    preds = func(self, list(input_data[:]), *args, **kwargs)

    for pred in preds:
      out_scores.append(pred.predicted_scores)

    out_tensor_scores = pb_utils.Tensor("softmax_predictions",
                                        np.asarray(out_scores, dtype=np.float32))
    inference_response = pb_utils.InferenceResponse(output_tensors=[out_tensor_scores])

    return inference_response

  return parse_predictions


def text_classifier(func: Callable):
  """
  Text classifier type output parser.
  """

  @wraps(func)
  def parse_predictions(self, input_data: np.ndarray, *args, **kwargs):
    """
    Format predictions and return clarifai compatible output.
    """
    out_scores = []
    input_data = [in_elem[0].decode() for in_elem in input_data]
    preds = func(self, input_data, *args, **kwargs)

    for pred in preds:
      out_scores.append(pred.predicted_scores)

    out_tensor_scores = pb_utils.Tensor("softmax_predictions",
                                        np.asarray(out_scores, dtype=np.float32))
    inference_response = pb_utils.InferenceResponse(output_tensors=[out_tensor_scores])

    return inference_response

  return parse_predictions


def text_to_text(func: Callable):
  """
  Text to text type output parser.
  Convert a sequence of text into another e.g. text generation,
  summarization or translation.
  """

  @wraps(func)
  def parse_predictions(self, input_data: np.ndarray, *args, **kwargs):
    """
    Format predictions and return clarifai compatible output.
    """
    out_text = []
    input_data = [in_elem[0].decode() for in_elem in input_data]
    preds = func(self, input_data, *args, **kwargs)

    for pred in preds:
      out_text.append(pred.predicted_text)

    out_text_tensor = pb_utils.Tensor("text", np.asarray(out_text, dtype=object))
    inference_response = pb_utils.InferenceResponse(output_tensors=[out_text_tensor])

    return inference_response

  return parse_predictions


def text_embedder(func: Callable):
  """
  Text embedder type output parser.
  Generates embeddings for an input text.
  """

  @wraps(func)
  def parse_predictions(self, input_data: np.ndarray, *args, **kwargs):
    """
    Format predictions and return clarifai compatible output.
    """
    out_embeddings = []
    input_data = [in_elem[0].decode() for in_elem in input_data]
    preds = func(self, input_data, *args, **kwargs)

    for pred in preds:
      out_embeddings.append(pred.embedding_vector)

    out_embed_tensor = pb_utils.Tensor("embeddings", np.asarray(out_embeddings, dtype=np.float32))
    inference_response = pb_utils.InferenceResponse(output_tensors=[out_embed_tensor])

    return inference_response

  return parse_predictions


def visual_embedder(func: Callable):
  """
  Visual embedder type output parser.
  Generates embeddings for an input image.
  """

  @wraps(func)
  def parse_predictions(self, input_data: np.ndarray, *args, **kwargs):
    """
    Format predictions and return clarifai compatible output.
    """
    out_embeddings = []
    # input_data passed as list of images
    preds = func(self, list(input_data[:]), *args, **kwargs)

    for pred in preds:
      out_embeddings.append(pred.embedding_vector)

    out_embed_tensor = pb_utils.Tensor("embeddings", np.asarray(out_embeddings, dtype=np.float32))
    inference_response = pb_utils.InferenceResponse(output_tensors=[out_embed_tensor])

    return inference_response

  return parse_predictions


def visual_segmenter(func: Callable):
  """
  Visual segmenter type output parser.
  """

  @wraps(func)
  def parse_predictions(self, input_data: np.ndarray, *args, **kwargs):
    """
    Format predictions and return clarifai compatible output.
    """
    masks = []
    # input_data passed as list of images
    preds = func(self, list(input_data[:]), *args, **kwargs)

    for pred in preds:
      masks.append(pred.predicted_mask)

    out_mask_tensor = pb_utils.Tensor("predicted_mask", np.asarray(masks, dtype=np.int64))
    inference_response = pb_utils.InferenceResponse(output_tensors=[out_mask_tensor])

    return inference_response

  return parse_predictions


def text_to_image(func: Callable):
  """
  Text to image type output parser.
  """

  @wraps(func)
  def parse_predictions(self, input_data: np.ndarray, *args, **kwargs):
    """
    Format predictions and return clarifai compatible output.
    """
    gen_images = []
    input_data = [in_elem[0].decode() for in_elem in input_data]
    preds = func(self, input_data, *args, **kwargs)

    for pred in preds:
      gen_images.append(pred.image)

    out_image_tensor = pb_utils.Tensor("image", np.asarray(gen_images, dtype=np.uint8))
    inference_response = pb_utils.InferenceResponse(output_tensors=[out_image_tensor])

    return inference_response

  return parse_predictions


def multimodal_embedder(func: Callable):
  """
  Multimodal embedder type output parser.
  Generates embeddings for image or text input.
  """

  @wraps(func)
  def parse_predictions(self, input_data: Dict[str, np.ndarray], *args, **kwargs):
    """
    Format predictions and return clarifai compatible output.
    """
    out_embeddings = []
    model_input_data = []
    for group in zip_longest(*input_data.values()):
      _input_data = dict(zip(input_data, group))
      for k, v in _input_data.items():
        # decode np.object to string
        if isinstance(v, np.ndarray) and v.dtype == np.object_:
          _input_data.update({k: v[0].decode()})
        model_input_data.append(_input_data)

    preds = func(self, model_input_data, *args, **kwargs)
    for pred in preds:
      out_embeddings.append(pred.embedding_vector)

    out_embed_tensor = pb_utils.Tensor("embeddings", np.asarray(out_embeddings, dtype=np.float32))
    inference_response = pb_utils.InferenceResponse(output_tensors=[out_embed_tensor])

    return inference_response

  return parse_predictions



import os

MAX_HW_DIM = 1024
IMAGE_TENSOR_NAME = "image"
TEXT_TENSOR_NAME = "text"

BUILT_MODEL_EXT = ".clarifai"

USER_CACHE_DIR = os.path.join(os.path.expanduser("~"), ".cache")
CLARIFAI_HOME = os.path.expanduser(
    os.getenv(
        "CLARIFAI_HOME",
        os.path.join(os.getenv("XDG_CACHE_HOME", USER_CACHE_DIR), "clarifai"),
    ))
os.makedirs(CLARIFAI_HOME, exist_ok=True)
CLARIFAI_PAT_PATH = os.path.join(CLARIFAI_HOME, "pat")

CLARIFAI_EXAMPLES_REPO = "https://github.com/Clarifai/examples.git"
repo_name = CLARIFAI_EXAMPLES_REPO.split("/")[-1].replace(".git", "")
CLARIFAI_EXAMPLES_REPO_PATH = os.path.join(CLARIFAI_HOME, repo_name)
MODEL_UPLOAD_EXAMPLE_FOLDER = "model_upload"



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Model Config classes."""
from __future__ import annotations  # isort: skip

from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, List, Union

from ...constants import IMAGE_TENSOR_NAME, MAX_HW_DIM


### Triton Model Config classes.###
@dataclass
class DType:
  """
  Triton Model Config data types.
  """
  # https://github.com/triton-inference-server/common/blob/main/protobuf/model_config.proto
  TYPE_UINT8: int = 2
  TYPE_INT8: int = 6
  TYPE_INT16: int = 7
  TYPE_INT32: int = 8
  TYPE_INT64: int = 9
  TYPE_FP16: int = 10
  TYPE_FP32: int = 11
  TYPE_STRING: int = 13
  KIND_GPU: int = 1
  KIND_CPU: int = 2


@dataclass
class InputConfig:
  """
  Triton Input definition.
  Params:
  -------
  name: input name
  data_type: input data type
  dims: Pre-defined input data shape(s).

  Returns:
  --------
  InputConfig
  """
  name: str
  data_type: int
  dims: List = field(default_factory=list)
  optional: bool = False


@dataclass
class OutputConfig:
  """
  Triton Output definition.
  Params:
  -------
  name: output name
  data_type: output data type
  dims: Pre-defined output data shape(s).
  labels (bool): If labels file is required for inference.

  Returns:
  --------
  OutputConfig
  """
  name: str
  data_type: int
  dims: List = field(default_factory=list)
  label_filename: str = ""


@dataclass
class Device:
  """
  Triton instance_group.
  Define the type of inference device and number of devices to use.
  Params:
  -------
  count: number of devices
  use_gpu: whether to use cpu or gpu.

  Returns:
  --------
  Device object
  """
  count: int = 1
  use_gpu: bool = True

  def __post_init__(self):
    if self.use_gpu:
      self.kind: str = DType.KIND_GPU
    else:
      self.kind: str = DType.KIND_CPU


@dataclass
class DynamicBatching:
  """
  Triton dynamic_batching config.
  Params:
  -------
  preferred_batch_size: batch size
  max_queue_delay_microseconds: max queue delay for a request batch

  Returns:
  --------
  DynamicBatching object
  """
  #preferred_batch_size: List[int] = [1] # recommended not to set
  max_queue_delay_microseconds: int = 500


@dataclass
class TritonModelConfig:
  """
  Triton Model Config base.
  Params:
  -------
  name: triton inference model name
  input: a list of an InputConfig field
  output: a list of OutputConfig fields/dicts
  instance_group: Device. see Device
  dynamic_batching: Triton dynamic batching settings.
  max_batch_size: max request batch size
  backend: Triton Python Backend. Constant
  image_shape: List of Height and Width of input image. *

  (*): This attribute won't be serialized in config.pbtxt

  Returns:
  --------
  TritonModelConfig
  """
  #model_type: str
  model_name: str = ""
  model_version: str = "1"
  input: List[InputConfig] = field(default_factory=list)
  output: List[OutputConfig] = field(default_factory=list)
  instance_group: Device = field(default_factory=Device)
  dynamic_batching: DynamicBatching = field(default_factory=DynamicBatching)
  max_batch_size: int = 1
  backend: str = "python"
  image_shape: tuple[Union[int, float], Union[int, float]] = field(
      default_factory=lambda: [-1, -1])  #(H, W)

  def __setattr__(self, __name: str, __value: Any) -> None:
    if __name == "image_shape":
      __value = self._check_and_assign_image_shape_value(__value)

    super().__setattr__(__name, __value)

  def _check_and_assign_image_shape_value(self, value):
    _has_image = False
    for each in self.input:
      if IMAGE_TENSOR_NAME in each.name:
        _has_image = True
        if len(value) != 2:
          raise ValueError(
              f"image_shape takes 2 values, Height and Width. Got {len(value)} values instead.")
        if value[0] > MAX_HW_DIM or value[1] > MAX_HW_DIM:
          raise ValueError(
              f"H and W each have a maximum value of {MAX_HW_DIM}. Got H: {value[0]}, W: {value[1]}"
          )
        image_dims = deepcopy(value)
        image_dims.append(3)  # add channel dim
        each.dims = image_dims

    if not _has_image and self.input:
      return [-1, -1]
    else:
      return value



import json
from dataclasses import asdict, dataclass, field
from typing import Any, List


@dataclass(frozen=True)
class InferParamType:
  BOOL: int = 1
  STRING: int = 2
  NUMBER: int = 3
  ENCRYPTED_STRING: int = 21


@dataclass
class InferParam:
  path: str
  field_type: InferParamType = field(default_factory=InferParamType)
  default_value: Any = None
  description: str = ""

  def __post_init__(self):
    assert self.path.isidentifier(
    ), f"`path` must be valid for creating python variable, got {self.path}"
    if self.default_value is not None:
      self.validate_type(self.default_value)

  def validate_type(self, value):
    if self.field_type == InferParamType.BOOL:
      assert isinstance(value, bool), f"`field_type` is `BOOL` (bool), however got {type(value)}"
    elif self.field_type == InferParamType.NUMBER:
      assert isinstance(value, float) or isinstance(
          value, int), f"`field_type` is `NUMBER` (float or int), however got {type(value)}"
    else:
      assert isinstance(
          value,
          str), f"`field_type` is `STRING` or `ENCRYPTED_STRING` (str), however got {type(value)}"

  def __setattr__(self, __name: str, __value: Any) -> None:
    if __name == "default_value":
      self.validate_type(__value)
    super().__setattr__(__name, __value)

  def todict(self):
    return {k: v for k, v in asdict(self).items()}


@dataclass
class InferParamManager:
  json_path: str = ""
  params: List[InferParam] = field(default_factory=list)
  _dict_params: dict = field(init=False)

  @classmethod
  def from_kwargs(cls, **kwargs):
    params = list()
    for k, v in kwargs.items():
      if isinstance(v, str) and k.startswith("_"):
        _type = InferParamType.ENCRYPTED_STRING
      elif isinstance(v, str):
        _type = InferParamType.STRING
      elif isinstance(v, bool):
        _type = InferParamType.BOOL
      elif isinstance(v, float) or isinstance(v, int):
        _type = InferParamType.NUMBER
      else:
        raise TypeError(f"Unsupported type {type(v)} of argument {k}, support {InferParamType}")
      param = InferParam(path=k, field_type=_type, default_value=v, description=k)
      params.append(param)

    return cls(params=params)

  def __post_init__(self):
    #assert self.params == [] or self.json_path, "`json_path` or `params` must be set"
    self._dict_params = dict()
    if self.params == [] and self.json_path:
      with open(self.json_path, "r") as fp:
        objs = json.load(fp)
        objs = objs if isinstance(objs, list) else [objs]
      self.params = [InferParam(**obj) for obj in objs]
    for param in self.params:
      self._dict_params.update({param.path: param})

  def get_list_params(self):
    list_params = []
    for each in self.params:
      list_params.append(each.todict())
    return list_params

  def export(self, path: str):
    list_params = self.get_list_params()
    with open(path, "w") as fp:
      json.dump(list_params, fp, indent=2)

  def validate(self, **kwargs) -> dict:
    output_kwargs = {k: v.default_value for k, v in self._dict_params.items()}
    assert kwargs == {} or self.params != [], "kwargs are rejected since `params` is empty"

    for key, value in kwargs.items():
      assert key in self._dict_params, f"param `{key}` is not in setting: {list(self._dict_params.keys())}"
      if key in self._dict_params:
        self._dict_params[key].validate_type(value)
      output_kwargs.update({key: value})
    return output_kwargs


def is_number(v: str):
  try:
    _ = float(v)
    return True
  except ValueError:
    return False


def str_to_number(v: str):
  try:
    return int(v)
  except ValueError:
    return float(v)


def parse_req_parameters(req_params: str):
  req_params = json.loads(req_params)
  for k, v in req_params.items():
    if isinstance(v, str):
      if is_number(v):
        v = str_to_number(v)
    req_params.update({k: v})

  return req_params



import os

from clarifai.models.model_serving.constants import CLARIFAI_PAT_PATH
from clarifai.utils.constants import CLARIFAI_PAT_ENV_VAR


def _persist_pat(pat: str):
  """ Write down pat to CLARIFAI_PAT_PATH """
  with open(CLARIFAI_PAT_PATH, "w") as f:
    f.write(pat)


def _read_pat():
  if not os.path.exists(CLARIFAI_PAT_PATH) and not os.environ.get(CLARIFAI_PAT_ENV_VAR, ""):
    return None
  if os.path.exists(CLARIFAI_PAT_PATH):
    with open(CLARIFAI_PAT_PATH, "r") as f:
      return f.read().replace("\n", "").replace("\r", "").strip()
  elif os.environ.get(CLARIFAI_PAT_ENV_VAR):
    return os.environ.get(CLARIFAI_PAT_ENV_VAR)
  else:
    raise ValueError(
        f"PAT not found, please run `clarifai login` to persist your PAT or set it as an environment variable under the name '{CLARIFAI_PAT_ENV_VAR}'"
    )


def login(pat=None):
  """ if pat provided, set pat to CLARIFAI_PAT otherwise read pat from file"""
  pat = pat or _read_pat()
  assert pat, Exception("PAT is not found, please run `clarifai login` to persist your PAT")
  os.environ["CLARIFAI_PAT"] = pat



# User model inference script.

import os
from pathlib import Path
from typing import Dict, Union
from clarifai.models.model_serving.model_config import *  # noqa


class InferenceModel():
  """User model inference class."""

  def __init__(self) -> None:
    """
    Load inference time artifacts that are called frequently .e.g. models, tokenizers, etc.
    in this method so they are loaded only once for faster inference.
    """
    # current directory
    self.base_path: Path = os.path.dirname(__file__)

  def predict(self,
              input_data: list,
              inference_parameters: Dict[str, Union[bool, str, float, int]] = {}) -> list:
    """predict_docstring
    """

    raise NotImplementedError()



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Commandline interface for model upload utils."""
import argparse
import os
import subprocess

from clarifai.models.model_serving.model_config import get_model_config, load_user_config
from clarifai.models.model_serving.model_config.inference_parameter import InferParamManager

from ..constants import BUILT_MODEL_EXT
from ..utils import login
from .base import BaseClarifaiCli


class UploadCli(BaseClarifaiCli):

  @staticmethod
  def register(parser: argparse._SubParsersAction):
    creator_parser = parser.add_parser("upload", help="Upload component to Clarifai platform")
    sub_creator_parser = creator_parser.add_subparsers()

    UploadModelSubCli.register(sub_creator_parser)

    creator_parser.set_defaults(func=UploadCli)


class UploadModelSubCli(BaseClarifaiCli):

  @staticmethod
  def register(parser: argparse._SubParsersAction):
    upload_parser = parser.add_parser("model", help="Upload Clarifai model")
    upload_parser.add_argument(
        "path",
        type=str,
        nargs='?',
        help=
        "Path to working dir to get clarifai_config.yaml or path to yaml. Default is current directory",
        default=".")
    upload_parser.add_argument(
        "--url", type=str, required=False, help="Direct download url of zip file", default=None)
    upload_parser.add_argument(
        "--file", type=str, required=False, help="Local built file", default=None)
    upload_parser.add_argument("--id", type=str, required=False, help="Model ID")
    upload_parser.add_argument(
        "--user-app",
        type=str,
        required=False,
        help="User ID and App ID separated by '/', e.g., <user_id>/<app_id>")
    upload_parser.add_argument(
        "--no-test",
        action="store_true",
        help="Trigger this flag to skip testing before uploading")
    upload_parser.add_argument(
        "--no-resume",
        action="store_true",
        help="Trigger this flag to not resume uploading local file")
    upload_parser.add_argument(
        "--update-version",
        action="store_true",
        required=False,
        help="Update exist model with new version")

    upload_parser.set_defaults(func=UploadModelSubCli)

  def __init__(self, args: argparse.Namespace) -> None:
    self.no_test = args.no_test
    self.no_resume = args.no_resume

    working_dir_or_config = args.path
    # if input a config file, then not running test
    if working_dir_or_config.endswith(".yaml"):
      # to folder
      working_dir_or_config = os.path.split(working_dir_or_config)[0]
      config_yaml_path = working_dir_or_config
      self.test_path = None
      self.no_test = True
    # if it is a directory - working dir then it must contain config and test
    else:
      config_yaml_path = os.path.join(working_dir_or_config, "clarifai_config.yaml")
      self.test_path = os.path.join(working_dir_or_config, "test.py")

    assert os.path.exists(config_yaml_path), FileNotFoundError(
        f"`{config_yaml_path}` does not exist")
    self.config = load_user_config(cfg_path=config_yaml_path)

    self.file = args.file
    self.url = args.url
    if self.file:
      assert not self.url, ValueError("Expected either file or url, not both.")
      assert os.path.exists(self.file), FileNotFoundError
    elif self.url:
      if len(self.url.split(":")) == 1:
        # if URL has no scheme, default to https
        self.url = f"https://{self.url}"
      assert self.url.startswith("http") or self.url.startswith("https") or self.url.startswith(
          "s3"
      ), f"Invalid URL scheme, supported schemes are 'http', 'https', or 's3'. Got {self.url}"
      self.file = None
    else:
      for _fname in os.listdir(working_dir_or_config):
        if _fname.endswith(BUILT_MODEL_EXT):
          self.file = os.path.join(working_dir_or_config, _fname)
          break
      assert self.file, ValueError(
          f"Not using url/file but also not found built file with extension {BUILT_MODEL_EXT}")

    self.user_id, self.app_id = "", ""
    user_app = args.user_app
    self.url: str = args.url
    self.update_version = args.update_version

    clarifai_cfg = self.config.clarifai_model
    self.url: str = args.url
    self.id = args.id or clarifai_cfg.clarifai_model_id
    self.type = clarifai_cfg.type
    self.desc = clarifai_cfg.description
    self.infer_param = clarifai_cfg.inference_parameters
    user_app = user_app or clarifai_cfg.clarifai_user_app_id

    if user_app:
      user_app = user_app.split('/')
      assert len(
          user_app
      ) == 2, f"id must be combination of user_id and app_id separated by `/`, e.g. <user_id>/<app_id>. Got {args.id}"
      self.user_id, self.app_id = user_app

    login()

  def run(self):
    from clarifai.client import App, Model

    # Run test before uploading
    if not self.no_test:
      assert os.path.exists(self.test_path), FileNotFoundError(f"Not found {self.test_path}")
      result = subprocess.run(f"pytest -s --log-level=INFO {self.test_path}", shell=True)
      assert result.returncode == 0, "Test has failed. Please make sure no error exists in your code."

    clarifai_key_map = get_model_config(model_type=self.type).clarifai_model.field_maps
    # inference parameters
    inference_parameters = None
    if isinstance(self.infer_param, str) and os.path.isfile(self.infer_param):
      inference_parameters = InferParamManager(json_path=self.infer_param).get_list_params()
    inputs = clarifai_key_map.input_fields_map
    outputs = clarifai_key_map.output_fields_map

    # if updating new version of existing model
    def update_version():
      model = Model(model_id=self.id, app_id=self.app_id)
      if self.url:
        model.create_version_by_url(
            url=self.url,
            input_field_maps=inputs,
            output_field_maps=outputs,
            inference_parameter_configs=inference_parameters,
            description=self.desc)
      elif self.file:
        model.create_version_by_file(
            file_path=self.file,
            input_field_maps=inputs,
            output_field_maps=outputs,
            inference_parameter_configs=inference_parameters,
            no_resume=self.no_resume,
            description=self.desc)
      else:
        raise ValueError

    if self.update_version:
      update_version()
    else:
      # creating new model
      _ = App(app_id=self.app_id).create_model(self.id, model_type_id=self.type)
      update_version()



from typing import Dict, Iterable, List, TypedDict, Union

import numpy as np

from ..constants import IMAGE_TENSOR_NAME, TEXT_TENSOR_NAME
from .config import ModelConfigClass, ModelTypes, get_model_config
from .output import (ClassifierOutput, EmbeddingOutput, ImageOutput, MasksOutput, TextOutput,
                     VisualDetectorOutput)
from .triton import wrappers as triton_wrapper


class _TypeCheckModelOutput(type):

  def __new__(cls, name, bases, attrs):
    """
    Override child `predict` function with parent._output_type_check(child.predict).
    Aim to check if child.predict returns valid output type
    """

    def wrap_function(fn_name, base, base_fn, other_fn):

      def new_fn(_self,
                 input_data,
                 inference_parameters: Dict[str, Union[bool, str, float, int]] = {}):
        # Run child class
        out = other_fn(_self, input_data, inference_parameters=inference_parameters)
        # Run type check
        return base_fn(base, input_data, out)

      new_fn.__name__ = "wrapped_%s" % fn_name
      new_fn.__doc__ = other_fn.__doc__
      return new_fn

    if name != "_BaseClarifaiModel":
      attrs["predict"] = wrap_function("predict", bases[0],
                                       getattr(bases[0], "_output_type_check", lambda: None),
                                       attrs.setdefault("predict", lambda: None))

    return type.__new__(cls, name, bases, attrs)


class _BaseClarifaiModel(metaclass=_TypeCheckModelOutput):
  _config: ModelConfigClass = None

  @property
  def config(self):
    return self._config

  def _output_type_check(self, input, output):
    output_type = self._config.clarifai_model.output_type
    if isinstance(output, Iterable):
      assert all(
          each.__class__.__name__ == output_type for each in output
      ), f"Expected output is iteration of `{output_type}` type, got iteration `{output}`"
      assert len(output) == len(
          input
      ), f"Input length and output length must be equal, but got input length of {len(input)} and output length of {len(output)}"
    else:
      raise ValueError(f"Expected output is iteration of `{output_type}` type, got `{output}`")
    return output

  def predict(self,
              input_data: Union[List[np.ndarray], Dict[str, List[np.ndarray]]],
              inference_parameters: Dict[str, Union[bool, str, float, int]] = {}) -> Iterable:
    """
    Prediction method.

    Args:
    -----
    - input_data: A list of input data item to predict on. The type depends on model input type:
      * `image`: List[np.ndarray]
      * `text`: List[str]
      * `multimodal`:
        input_data is list of dict where key is input type name e.i. `image`, `text` and value is list.
        {"image": List[np.ndarray], "text": List[str]}

    - inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters.

    Returns:
    --------
      List of one of the `clarifai.models.model_serving.model_config.output` types. Refer to the README/docs
    """
    raise NotImplementedError

  def _tritonserver_predict(self,
                            input_data,
                            inference_parameters: Dict[str, Union[bool, str, float, int]] = {}):
    """ This method is invoked within tritonserver, specifically in the model.py of the Python backend. Attempting to execute it outside of the triton environment will result in failure."""
    raise NotImplementedError


_MultiModalInputTypeDict = TypedDict("_MultiModalInputTypeDict", {
    IMAGE_TENSOR_NAME: np.ndarray,
    TEXT_TENSOR_NAME: str
})


class MultiModalEmbedder(_BaseClarifaiModel):
  _config: ModelConfigClass = get_model_config(ModelTypes.multimodal_embedder)

  def predict(self,
              input_data: List[_MultiModalInputTypeDict],
              inference_parameters: Dict[str, Union[bool, str, float, int]] = {}
             ) -> Iterable[EmbeddingOutput]:
    """ Custom prediction function for `multimodal-embedder` model.

    Args:
      input_data (List[_MultiModalInputTypeDict]): List of dict of key-value: `image`(np.ndarray) and `text` (str)
      inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters

    Returns:
      list of EmbeddingOutput
    """
    raise NotImplementedError

  @triton_wrapper.multimodal_embedder
  def _tritonserver_predict(self,
                            input_data,
                            inference_parameters: Dict[str, Union[bool, str, float, int]] = {}):
    """ This method is invoked within tritonserver, specifically in the model.py of the Python backend. Attempting to execute it outside of the triton environment will result in failure."""
    return self.predict(input_data, inference_parameters=inference_parameters)


class TextClassifier(_BaseClarifaiModel):
  _config: ModelConfigClass = get_model_config(ModelTypes.text_classifier)

  def predict(self,
              input_data: List[str],
              inference_parameters: Dict[str, Union[bool, str, float, int]] = {}
             ) -> Iterable[ClassifierOutput]:
    """ Custom prediction function for `text-classifier` model.

    Args:
      input_data (List[str]): List of text
      inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters

    Returns:
      list of ClassifierOutput
    """
    raise NotImplementedError

  @triton_wrapper.text_classifier
  def _tritonserver_predict(self,
                            input_data,
                            inference_parameters: Dict[str, Union[bool, str, float, int]] = {}):
    return self.predict(input_data, inference_parameters=inference_parameters)


class TextEmbedder(_BaseClarifaiModel):
  _config: ModelConfigClass = get_model_config(ModelTypes.text_embedder)

  def predict(self,
              input_data: List[str],
              inference_parameters: Dict[str, Union[bool, str, float, int]] = {}
             ) -> Iterable[EmbeddingOutput]:
    """ Custom prediction function for `text-embedder` model.

    Args:
      input_data (List[str]): List of text
      inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters

    Returns:
      list of EmbeddingOutput
    """
    raise NotImplementedError

  @triton_wrapper.text_embedder
  def _tritonserver_predict(self,
                            input_data,
                            inference_parameters: Dict[str, Union[bool, str, float, int]] = {}):
    return self.predict(input_data, inference_parameters=inference_parameters)


class TextToImage(_BaseClarifaiModel):
  _config: ModelConfigClass = get_model_config(ModelTypes.text_to_image)

  def predict(
      self,
      input_data: List[str],
      inference_parameters: Dict[str, Union[bool, str, float, int]] = {}) -> Iterable[ImageOutput]:
    """ Custom prediction function for `text-to-image` model.

    Args:
      input_data (List[str]): List of text
      inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters

    Returns:
      list of ImageOutput
    """
    raise NotImplementedError

  @triton_wrapper.text_to_image
  def _tritonserver_predict(self,
                            input_data,
                            inference_parameters: Dict[str, Union[bool, str, float, int]] = {}):
    return self.predict(input_data, inference_parameters=inference_parameters)


class TextToText(_BaseClarifaiModel):
  _config: ModelConfigClass = get_model_config(ModelTypes.text_to_text)

  def predict(
      self,
      input_data: List[str],
      inference_parameters: Dict[str, Union[bool, str, float, int]] = {}) -> Iterable[TextOutput]:
    """ Custom prediction function for `text-to-text` (also called as `text generation`) model.

    Args:
      input_data (List[str]): List of text
      inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters

    Returns:
      list of TextOutput
    """
    raise NotImplementedError

  @triton_wrapper.text_to_text
  def _tritonserver_predict(self,
                            input_data,
                            inference_parameters: Dict[str, Union[bool, str, float, int]] = {}):
    """ This method is invoked within tritonserver, specifically in the model.py of the Python backend. Attempting to execute it outside of the triton environment will result in failure."""
    return self.predict(input_data, inference_parameters=inference_parameters)


class VisualClassifier(_BaseClarifaiModel):
  _config: ModelConfigClass = get_model_config(ModelTypes.visual_classifier)

  def predict(self,
              input_data: List[np.ndarray],
              inference_parameters: Dict[str, Union[bool, str, float, int]] = {}
             ) -> Iterable[ClassifierOutput]:
    """ Custom prediction function for `visual-classifier` model.

    Args:
      input_data (List[np.ndarray]): List of image
      inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters

    Returns:
      list of ClassifierOutput
    """
    raise NotImplementedError

  @triton_wrapper.visual_classifier
  def _tritonserver_predict(self,
                            input_data,
                            inference_parameters: Dict[str, Union[bool, str, float, int]] = {}):
    """ This method is invoked within tritonserver, specifically in the model.py of the Python backend. Attempting to execute it outside of the triton environment will result in failure."""
    return self.predict(input_data, inference_parameters=inference_parameters)


class VisualDetector(_BaseClarifaiModel):
  _config: ModelConfigClass = get_model_config(ModelTypes.visual_detector)

  def predict(self,
              input_data: List[np.ndarray],
              inference_parameters: Dict[str, Union[bool, str, float, int]] = {}
             ) -> Iterable[VisualDetectorOutput]:
    """ Custom prediction function for `visual-detector` model.

    Args:
      input_data (List[np.ndarray]): List of image
      inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters

    Returns:
      list of VisualDetectorOutput
    """
    raise NotImplementedError

  @triton_wrapper.visual_detector
  def _tritonserver_predict(self,
                            input_data,
                            inference_parameters: Dict[str, Union[bool, str, float, int]] = {}):
    """ This method is invoked within tritonserver, specifically in the model.py of the Python backend. Attempting to execute it outside of the triton environment will result in failure."""
    return self.predict(input_data, inference_parameters=inference_parameters)

  @staticmethod
  def postprocess(width: int,
                  height: int,
                  labels: list,
                  scores: list,
                  xyxy_boxes: list,
                  max_bbox_count: int = 500) -> VisualDetectorOutput:
    """Convert detection output to Clarifai detector output format

    Args:
        width (int): image width
        height (int): image height
        labels (list): list of labels
        scores (list): list of scores
        xyxy_boxes (list): list of bounding boxes in x_min, y_min, x_max, y_max format
        max_bbox_count (int, optional): Maximum detection result. Defaults to 500.

    Returns:
        VisualDetectorOutput
    """
    assert len(labels) == len(scores) == len(
        xyxy_boxes
    ), f"Length of `labels`, `scores` and `bboxes` must be equal, got {len(labels)}, {len(scores)} and {len(xyxy_boxes)} "
    labels = [[each] for each in labels]
    scores = [[each] for each in scores]
    bboxes = [[x[1] / height, x[0] / width, x[3] / height, x[2] / width]
              for x in xyxy_boxes]  # normalize the bboxes to [0,1] and [y1 x1 y2 x2]
    bboxes = np.clip(bboxes, 0, 1.)
    if len(bboxes) != 0:
      bboxes = np.concatenate((bboxes, np.zeros((max_bbox_count - len(bboxes), 4))))
      scores = np.concatenate((scores, np.zeros((max_bbox_count - len(scores), 1))))
      labels = np.concatenate((labels, np.zeros((max_bbox_count - len(labels), 1),
                                                dtype=np.int32)))
    else:
      bboxes = np.zeros((max_bbox_count, 4), dtype=np.float32)
      scores = np.zeros((max_bbox_count, 1), dtype=np.float32)
      labels = np.zeros((max_bbox_count, 1), dtype=np.int32)

    output = VisualDetectorOutput(
        predicted_bboxes=bboxes, predicted_labels=labels, predicted_scores=scores)

    return output


class VisualEmbedder(_BaseClarifaiModel):
  _config: ModelConfigClass = get_model_config(ModelTypes.visual_embedder)

  def predict(self,
              input_data: List[np.ndarray],
              inference_parameters: Dict[str, Union[bool, str, float, int]] = {}
             ) -> Iterable[EmbeddingOutput]:
    """ Custom prediction function for `visual-embedder` model.

    Args:
      input_data (List[np.ndarray]): List of image
      inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters

    Returns:
      list of EmbeddingOutput
    """
    raise NotImplementedError

  @triton_wrapper.visual_embedder
  def _tritonserver_predict(self,
                            input_data,
                            inference_parameters: Dict[str, Union[bool, str, float, int]] = {}):
    """ This method is invoked within tritonserver, specifically in the model.py of the Python backend. Attempting to execute it outside of the triton environment will result in failure."""
    return self.predict(input_data, inference_parameters=inference_parameters)


class VisualSegmenter(_BaseClarifaiModel):
  _config: ModelConfigClass = get_model_config(ModelTypes.visual_segmenter)

  def predict(
      self,
      input_data: List[np.ndarray],
      inference_parameters: Dict[str, Union[bool, str, float, int]] = {}) -> Iterable[MasksOutput]:
    """ Custom prediction function for `visual-segmenter` model.

    Args:
      input_data (List[np.ndarray]): List of image
      inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters

    Returns:
      list of MasksOutput
    """
    raise NotImplementedError

  @triton_wrapper.visual_segmenter
  def _tritonserver_predict(self,
                            input_data,
                            inference_parameters: Dict[str, Union[bool, str, float, int]] = {}):
    """ This method is invoked within tritonserver, specifically in the model.py of the Python backend. Attempting to execute it outside of the triton environment will result in failure."""
    return self.predict(input_data, inference_parameters=inference_parameters)



from argparse import ArgumentParser

from .build import BuildCli
from .create import CreateCli
from .example_cli import ExampleCli
from .login import LoginCli
from .upload import UploadCli


def main():

  parser = ArgumentParser("clarifai")
  cmd_parser = parser.add_subparsers(help="Clarifai cli helpers")

  UploadCli.register(cmd_parser)
  CreateCli.register(cmd_parser)
  LoginCli.register(cmd_parser)
  ExampleCli.register(cmd_parser)
  BuildCli.register(cmd_parser)

  args = parser.parse_args()

  if not hasattr(args, "func"):
    parser.print_help()
    exit(1)

  # Run
  service = args.func(args)
  service.run()


if __name__ == "__main__":
  main()
